from . import aggregator as agg
from . import dimensions as dim
from . import plot
from . import util
from .dataset.imaging import MaskedImaging, SettingsMaskedImaging, SimulatorImaging
from .dataset.interferometer import (
    MaskedInterferometer,
    SettingsMaskedInterferometer,
    SimulatorInterferometer,
)
from autoarray import GridIterate
from autoarray import GridInterpolate
from autoarray import Mask2D
from autoarray import GridCoordinates
from autoarray import TransformerDFT
from autoarray import pix
from autoarray import reg
from autoarray import TransformerNUFFT
from autoarray import SettingsInversion
from autoarray import SettingsPixelization
from autoarray.structures.arrays import Array
from autoarray.structures.grids import Grid, GridRectangular, GridVoronoi
from autoarray.structures.kernel import Kernel
from autoarray.structures.visibilities import Visibilities
from autoarray.dataset.imaging import Imaging
from autoarray.dataset.interferometer import Interferometer
from autoarray.operators.convolver import Convolver
from autoconf import conf

from .fit.fit import FitImaging, FitInterferometer
from .galaxy.fit_galaxy import FitGalaxy
from .galaxy.galaxy import Galaxy, HyperGalaxy, Redshift
from .galaxy.galaxy_data import GalaxyData
from .galaxy.galaxy_model import GalaxyModel
from .galaxy.masked_galaxy_data import MaskedGalaxyDataset
from .hyper import hyper_data
from .pipeline.phase.abstract import phase
from .pipeline.phase.abstract.phase import AbstractPhase
from .pipeline.phase.extensions import CombinedHyperPhase
from .pipeline.phase.extensions import HyperGalaxyPhase
from .pipeline.phase.extensions.hyper_galaxy_phase import HyperGalaxyPhase
from .pipeline.phase.extensions.hyper_phase import HyperPhase
from .pipeline.phase.extensions.inversion_phase import InversionPhase
from .pipeline.phase.extensions.inversion_phase import ModelFixingHyperPhase
from .pipeline.phase.imaging.phase import PhaseImaging
from .pipeline.phase.interferometer.phase import PhaseInterferometer
from .pipeline.phase.phase_galaxy import PhaseGalaxy
from .pipeline.phase.settings import SettingsPhaseImaging
from .pipeline.phase.settings import SettingsPhaseInterferometer
from .pipeline.pipeline import PipelineDataset
from .pipeline.setup import (
    SetupPipeline,
    SetupHyper,
    SetupLightParametric,
    SetupLightInversion,
    SetupMassTotal,
    SetupMassLightDark,
    SetupSMBH,
)
from .plane.plane import Plane
from .profiles import (
    light_profiles as lp,
    mass_profiles as mp,
    light_and_mass_profiles as lmp,
)
from .util import convert

conf.instance.register(__file__)

__version__ = '0.15.1'



from os import path

import pytest
from matplotlib import pyplot

from autoconf import conf


class PlotPatch:
    def __init__(self):
        self.paths = []

    def __call__(self, path, *args, **kwargs):
        self.paths.append(path)


@pytest.fixture(name="plot_patch")
def make_plot_patch(monkeypatch):
    plot_patch = PlotPatch()
    monkeypatch.setattr(pyplot, "savefig", plot_patch)
    return plot_patch


directory = path.dirname(path.realpath(__file__))


@pytest.fixture(autouse=True)
def set_config_path():
    conf.instance = conf.Config.for_directory(directory)



from autoconf import conf
import autofit as af
import autogalaxy as ag

import pytest


class TestCase:
    def test_integration(self):
        # Create a mapper. This can be used to convert values output by a non linear optimiser into class instances.
        mapper = af.ModelMapper()

        # Create a model_galaxy prior for the source model_galaxy. Here we are describing only the light profile of
        # the source model_galaxy which comprises an elliptical exponential and elliptical sersic light profile.
        source_galaxy_prior = ag.GalaxyModel(
            redshift=ag.Redshift,
            light_profile_one=ag.lp.EllipticalExponential,
            light_profile_2=ag.lp.EllipticalSersic,
        )

        # Create a model_galaxy prior for the source model_galaxy. Here we are describing both the light and mass
        # profiles. We've also stipulated that the centres of any galaxies generated using the model_galaxy prior
        # should match.
        galaxy_galaxy_prior = ag.GalaxyModel(
            redshift=ag.Redshift,
            light_profile=ag.lp.EllipticalExponential,
            mass_profile=ag.mp.EllipticalExponential,
            align_centres=True,
        )

        mapper.galaxy_1 = source_galaxy_prior
        mapper.galaxy_0 = galaxy_galaxy_prior

        # Create a model instance. All the instances of the profile classes are created here. Normally we would do this
        # using the output of a non linear search but in this case we are using the median values from the priors.
        instance = mapper.instance_from_prior_medians()

        # Recover model_galaxy instances. We can pass the model instance to model_galaxy priors to recover a fully
        # constructed model_galaxy
        galaxy_1 = instance.galaxy_1
        galaxy_0 = instance.galaxy_0

        # Let's just check that worked
        assert len(galaxy_1.light_profiles) == 2
        assert len(galaxy_1.mass_profiles) == 0

        assert len(galaxy_0.light_profiles) == 1
        assert len(galaxy_0.mass_profiles) == 1

        assert galaxy_1.redshift == 1.5
        assert galaxy_0.redshift == 1.5



import autofit as af
from autogalaxy.pipeline.phase import abstract

from .hyper_phase import HyperPhase


# noinspection PyAbstractClass
class ModelFixingHyperPhase(HyperPhase):
    def __init__(
        self,
        phase: abstract.AbstractPhase,
        hyper_search,
        hyper_name: str,
        model_classes=tuple(),
    ):

        super().__init__(phase=phase, hyper_search=hyper_search, hyper_name=hyper_name)

        self.model_classes = model_classes

    def make_hyper_phase(self):
        return super().make_hyper_phase()

    def make_model(self, instance):
        return instance.as_model(self.model_classes)

    def run_hyper(
        self,
        dataset,
        results: af.ResultsCollection,
        info=None,
        pickle_files=None,
        **kwargs
    ):
        """
        Run the phase, overriding the search's model instance with one created to
        only fit pixelization hyperparameters.
        """

        self.results = results

        phase = self.make_hyper_phase()
        phase.model = self.make_model(results.last.instance)

        return phase.run(
            dataset,
            mask=results.last.mask,
            results=results,
            info=info,
            pickle_files=pickle_files,
        )


class InversionPhase(ModelFixingHyperPhase):
    def __init__(
        self,
        phase: abstract.AbstractPhase,
        hyper_search,
        model_classes=tuple(),
        inversion_pixels_fixed=None,
    ):
        super().__init__(
            phase=phase,
            hyper_search=hyper_search,
            model_classes=model_classes,
            hyper_name="inversion",
        )

        self.inversion_pixels_fixed = inversion_pixels_fixed

    def make_model(self, instance):
        model = instance.as_model(self.model_classes)

        # TODO : More checks here... Need Rich to build a more clever method.

        if self.inversion_pixels_fixed is not None and self.uses_cluster_inversion:
            if hasattr(model.galaxies, "source"):
                model.galaxies.source.pixelization.pixels = self.inversion_pixels_fixed
        return model



import time

import autogalaxy as ag

# Although we could test_autoarray the deflection angles without using an image (e.ag.mp. by just making a grid), we have chosen to
# set this test_autoarray up using an image and mask. This gives run-time numbers that can be easily related to an actual galaxy
# analysis

shape_2d = (301, 301)
pixel_scales = 0.05
sub_size = 4
radius = 4.0

print("sub grid size = " + str(sub_size))
print("circular mask radius = " + str(radius) + "\n")


mask = ag.Mask2D.circular(
    shape_2d=shape_2d, pixel_scales=pixel_scales, sub_size=sub_size, radius=radius
)

grid = ag.Grid.from_mask(mask=mask)

print("Number of points = " + str(grid.sub_shape_1d) + "\n")

### EllipticalIsothermal ###

mass_profile = ag.mp.EllipticalIsothermal(
    centre=(0.0, 0.0), elliptical_comps=(0.111111, 0.0), einstein_radius=1.0
)

image_plane_grid = ag.Grid.uniform(shape_2d=(500, 500), pixel_scales=0.02)

deflections = mass_profile.deflections_from_grid(grid=image_plane_grid)

deflections_y = ag.Array.manual_mask(
    array=deflections.in_1d[:, 0], mask=image_plane_grid.mask
)
deflections_x = ag.Array.manual_mask(
    array=deflections.in_1d[:, 1], mask=image_plane_grid.mask
)

input_deflections = ag.mp.InputDeflections(
    deflections_y=deflections_y,
    deflections_x=deflections_x,
    image_plane_grid=image_plane_grid,
    preload_grid=grid,
)

### Proifle ###

start = time.time()
input_deflections.deflections_from_grid(grid=grid)
diff = time.time() - start
print("InputDeflections time = {}".format(diff))



import autofit as af
import numpy as np
from astropy import cosmology as cosmo
from autoarray.structures import grids
from autogalaxy import dimensions as dim
from autogalaxy.profiles import geometry_profiles
from autogalaxy.profiles import mass_profiles as mp
from autogalaxy.util import convert
import typing

from scipy.interpolate import griddata
from autogalaxy import exc


class MassSheet(geometry_profiles.SphericalProfile, mp.MassProfile):
    @af.map_types
    def __init__(self, centre: dim.Position = (0.0, 0.0), kappa: float = 0.0):
        """
        Represents a mass-sheet

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        kappa : float
            The magnitude of the convergence of the mass-sheet.
        """
        super(MassSheet, self).__init__(centre=centre)
        self.kappa = kappa

    def convergence_func(self, grid_radius):
        return 0.0

    @grids.grid_like_to_structure
    def convergence_from_grid(self, grid):
        return np.full(shape=grid.shape[0], fill_value=self.kappa)

    @grids.grid_like_to_structure
    def potential_from_grid(self, grid):
        return np.zeros(shape=grid.shape[0])

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def deflections_from_grid(self, grid):
        grid_radii = self.grid_to_grid_radii(grid=grid)
        return self.grid_to_grid_cartesian(grid=grid, radius=self.kappa * grid_radii)

    @property
    def is_mass_sheet(self):
        return True


# noinspection PyAbstractClass
class ExternalShear(geometry_profiles.EllipticalProfile, mp.MassProfile):
    @af.map_types
    def __init__(self, elliptical_comps: typing.Tuple[float, float] = (0.0, 0.0)):
        """
        An `ExternalShear` term, to model the line-of-sight contribution of other galaxies / satellites.

        The shear angle phi is defined in the direction of stretching of the image. Therefore, if an object located \
        outside the lens is responsible for the shear, it will be offset 90 degrees from the value of phi.

        Parameters
        ----------
        magnitude : float
            The overall magnitude of the shear (gamma).
        phi : float
            The rotation axis of the shear.
        """

        super(ExternalShear, self).__init__(
            centre=(0.0, 0.0), elliptical_comps=elliptical_comps
        )

        magnitude, phi = convert.shear_magnitude_and_phi_from(
            elliptical_comps=elliptical_comps
        )

        self.magnitude = magnitude
        self.phi = phi

    def convergence_func(self, grid_radius):
        return 0.0

    def average_convergence_of_1_radius_in_units(
        self,
        unit_length="arcsec",
        redshift_profile=None,
        cosmology=cosmo.Planck15,
        **kwargs,
    ):
        return dim.Length(value=0.0, unit_length=self.unit_length)

    @grids.grid_like_to_structure
    def convergence_from_grid(self, grid):
        return np.zeros(shape=grid.shape[0])

    @grids.grid_like_to_structure
    def potential_from_grid(self, grid):
        return np.zeros(shape=grid.shape[0])

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def deflections_from_grid(self, grid):
        """
        Calculate the deflection angles at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : aa.Grid
            The grid of (y,x) arc-second coordinates the deflection angles are computed on.

        """
        deflection_y = -np.multiply(self.magnitude, grid[:, 0])
        deflection_x = np.multiply(self.magnitude, grid[:, 1])
        return self.rotate_grid_from_profile(np.vstack((deflection_y, deflection_x)).T)


class InputDeflections(mp.MassProfile):
    def __init__(
        self,
        deflections_y,
        deflections_x,
        image_plane_grid,
        preload_grid=None,
        preload_blurring_grid=None,
        #      normalization_scale: float = 1.0,
    ):
        """
        Represents a known deflection angle map (e.g. from an already performed lens model or particle simulation
        of a mass distribution) which can be used for model fitting.
        
        The image-plane grid of the delflection angles is used to align an input grid to the input deflections, so that
        a new deflection angle map can be computed via interpolation using the scipy.interpolate.griddata method.

        A normalization scale can be included, which scales the overall normalization of the deflection angle map
        interpolated by a multiplicative factor.

        Parameters
        ----------
        deflections_y : aa.Array
            The input array of the y components of the deflection angles.
        deflections_x : aa.Array
            The input array of the x components of the deflection angles.
        image_plane_grid : aa.Grid
            The image-plane grid from which the deflection angles are defined.
        grid_interp : aa.Grid
            The grid that interpolated quantities are computed on. If this is input in advance, the interpolation
            weights can be precomputed to speed up the calculation time.
        normalization_scale : float
            The calculated deflection angles are multiplied by this factor scaling their values up and doown.
        """
        super().__init__()

        self.deflections_y = deflections_y
        self.deflections_x = deflections_x

        self.image_plane_grid = image_plane_grid

        self.centre = image_plane_grid.origin

        self.preload_grid = preload_grid
        self.preload_deflections = None
        self.preload_blurring_grid = preload_blurring_grid
        self.preload_blurring_deflections = None

        if self.preload_grid is not None:
            self.normalization_scale = 1.0
            self.preload_deflections = self.deflections_from_grid(grid=preload_grid)

        if self.preload_blurring_grid is not None:
            self.normalization_scale = 1.0
            self.preload_blurring_deflections = self.deflections_from_grid(
                grid=preload_blurring_grid
            )

        self.normalization_scale = 1.0  # normalization_scale

    @grids.grid_like_to_structure
    def convergence_from_grid(self, grid):
        return self.convergence_via_jacobian_from_grid(grid=grid)

    @grids.grid_like_to_structure
    def potential_from_grid(self, grid):
        return np.zeros(shape=grid.shape[0])

    @grids.grid_like_to_structure
    def deflections_from_grid(self, grid):

        if self.preload_grid is not None and self.preload_deflections is not None:

            if grid.sub_shape_1d == self.preload_grid.sub_shape_1d:
                if np.allclose(grid, self.preload_grid, 1e-8):
                    return self.normalization_scale * self.preload_deflections

        if (
            self.preload_blurring_grid is not None
            and self.preload_blurring_deflections is not None
        ):

            if grid.sub_shape_1d == self.preload_blurring_grid.sub_shape_1d:
                if np.allclose(grid, self.preload_blurring_grid, 1e-8):
                    return self.normalization_scale * self.preload_blurring_deflections

        deflections_y = self.normalization_scale * griddata(
            points=self.image_plane_grid, values=self.deflections_y, xi=grid
        )
        deflections_x = self.normalization_scale * griddata(
            points=self.image_plane_grid, values=self.deflections_x, xi=grid
        )

        if np.isnan(deflections_y).any() or np.isnan(deflections_x).any():
            raise exc.ProfileException(
                "The grid input into the DefectionsInput.deflections_from_grid() method has (y,x)"
                "coodinates extending beyond the input image_plane_grid."
                ""
                "Update the image_plane_grid to include deflection angles reaching to larger"
                "radii or reduce the input grid. "
            )

        return np.stack((deflections_y, deflections_x), axis=-1)



import math

from astropy import constants

import autofit as af
import autogalaxy as ag
from autoarray.mock import *
from autoarray.structures import grids
from autofit.non_linear.mock.mock_search import MockSamples
from autogalaxy.pipeline.phase.dataset import PhaseDataset
from autogalaxy.plot.lensing_plotters import Include
from autofit.mock import MockSearch


# MockProfiles #


class MockLightProfile(ag.lp.LightProfile):
    def __init__(self, value, size=1):
        self.value = value
        self.size = size

    def image_from_grid(self, grid):
        return np.array(self.size * [self.value])


class MockMassProfile:
    def __init__(self, value):
        self.value = value

    def surface_density_from_grid(self, grid):
        return np.array([self.value])

    def potential_from_grid(self, grid):
        return np.array([self.value])

    def deflections_from_grid(self, grid):
        return np.array([self.value, self.value])


# Mock Galaxy #


class MockGalaxy:
    def __init__(self, value, shape=1):
        self.value = value
        self.shape = shape

    @grids.grid_like_to_structure
    def image_from_grid(self, grid):
        return np.full(shape=self.shape, fill_value=self.value)

    @grids.grid_like_to_structure
    def convergence_from_grid(self, grid):
        return np.full(shape=self.shape, fill_value=self.value)

    @grids.grid_like_to_structure
    def potential_from_grid(self, grid):
        return np.full(shape=self.shape, fill_value=self.value)

    @grids.grid_like_to_structure
    def deflections_from_grid(self, grid):
        return np.full(shape=(self.shape, 2), fill_value=self.value)


class MockHyperGalaxy:
    def __init__(self, contribution_factor=0.0, noise_factor=0.0, noise_power=1.0):
        self.contribution_factor = contribution_factor
        self.noise_factor = noise_factor
        self.noise_power = noise_power

    def contributions_from_model_image_and_galaxy_image(
        self, model_image, galaxy_image, minimum_value
    ):
        contributions = galaxy_image / (model_image + self.contribution_factor)
        contributions = contributions / np.max(contributions)
        contributions[contributions < minimum_value] = 0.0
        return contributions

    def hyper_noise_from_contributions(self, noise_map, contributions):
        return self.noise_factor * (noise_map * contributions) ** self.noise_power


# Mock Cosmology #


class Value:
    def __init__(self, value):
        self.value = value

    def to(self, *args, **kwargs):
        return Value(value=self.value)


class MockCosmology:
    def __init__(
        self,
        arcsec_per_kpc=0.5,
        kpc_per_arcsec=2.0,
        critical_surface_density=2.0,
        cosmic_average_density=2.0,
    ):
        self.arcsec_per_kpc = arcsec_per_kpc
        self.kpc_per_arcsec = kpc_per_arcsec
        self.critical_surface_density = critical_surface_density
        self.cosmic_average_density = cosmic_average_density

    def arcsec_per_kpc_proper(self, z):
        return Value(value=self.arcsec_per_kpc)

    def kpc_per_arcsec_proper(self, z):
        return Value(value=self.kpc_per_arcsec)

    def angular_diameter_distance(self, z):
        return Value(value=1.0)

    def angular_diameter_distance_z1z2(self, z1, z2):
        const = constants.c.to("kpc / s") ** 2.0 / (
            4 * math.pi * constants.G.to("kpc3 / (solMass s2)")
        )
        return Value(value=self.critical_surface_density * const.value)

    def critical_density(self, z):
        return Value(value=self.cosmic_average_density)


# Mock Pipeline / Phase #


class MockResult(af.MockResult):
    def __init__(
        self,
        samples=None,
        instance=None,
        model=None,
        analysis=None,
        search=None,
        mask=None,
        model_image=None,
        hyper_galaxy_image_path_dict=None,
        hyper_model_image=None,
        hyper_galaxy_visibilities_path_dict=None,
        hyper_model_visibilities=None,
        pixelization=None,
        use_as_hyper_dataset=False,
    ):
        super().__init__(
            samples=samples,
            instance=instance,
            model=model,
            analysis=analysis,
            search=search,
        )

        self.mask = mask
        self.hyper_galaxy_image_path_dict = hyper_galaxy_image_path_dict
        self.hyper_model_image = hyper_model_image
        self.hyper_galaxy_visibilities_path_dict = hyper_galaxy_visibilities_path_dict
        self.hyper_model_visibilities = hyper_model_visibilities
        self.model_image = model_image
        self.unmasked_model_image = model_image
        self.pixelization = pixelization
        self.use_as_hyper_dataset = use_as_hyper_dataset

        self.max_log_likelihood_plane = ag.Plane(galaxies=[ag.Galaxy(redshift=0.5)])

    @property
    def last(self):
        return self


class MockResults(af.ResultsCollection):
    def __init__(
        self,
        samples=None,
        instance=None,
        model=None,
        analysis=None,
        search=None,
        mask=None,
        model_image=None,
        hyper_galaxy_image_path_dict=None,
        hyper_model_image=None,
        hyper_galaxy_visibilities_path_dict=None,
        hyper_model_visibilities=None,
        pixelization=None,
        use_as_hyper_dataset=False,
    ):
        """
        A collection of results from previous phases. Results can be obtained using an index or the name of the phase
        from whence they came.
        """

        super().__init__()

        result = MockResult(
            samples=samples,
            instance=instance,
            model=model,
            analysis=analysis,
            search=search,
            mask=mask,
            model_image=model_image,
            hyper_galaxy_image_path_dict=hyper_galaxy_image_path_dict,
            hyper_model_image=hyper_model_image,
            hyper_galaxy_visibilities_path_dict=hyper_galaxy_visibilities_path_dict,
            hyper_model_visibilities=hyper_model_visibilities,
            pixelization=pixelization,
            use_as_hyper_dataset=use_as_hyper_dataset,
        )

        self.__result_list = [result]

    @property
    def last(self):
        """
        The result of the last phase
        """
        if len(self.__result_list) > 0:
            return self.__result_list[-1]
        return None

    def __getitem__(self, item):
        """
        Get the result of a previous phase by index

        Parameters
        ----------
        item: int
            The index of the result

        Returns
        -------
        result: Result
            The result of a previous phase
        """
        return self.__result_list[item]

    def __len__(self):
        return len(self.__result_list)


def make_masked_imaging_7x7():
    return ag.MaskedImaging(
        imaging=make_imaging_7x7(),
        mask=make_sub_mask_7x7(),
        settings=ag.SettingsMaskedImaging(sub_size=1),
    )


def make_masked_interferometer_7():
    return ag.MaskedInterferometer(
        interferometer=make_interferometer_7(),
        visibilities_mask=make_visibilities_mask_7x2(),
        real_space_mask=make_mask_7x7(),
        settings=ag.SettingsMaskedInterferometer(
            sub_size=1, transformer_class=ag.TransformerDFT
        ),
    )


def make_masked_interferometer_7_lop():
    return ag.MaskedInterferometer(
        interferometer=make_interferometer_7(),
        visibilities_mask=make_visibilities_mask_7x2(),
        real_space_mask=make_mask_7x7(),
        settings=ag.SettingsMaskedInterferometer(transformer_class=ag.TransformerNUFFT),
    )


#
# MODEL #
#

# PROFILES #


def make_lp_0():
    # noinspection PyTypeChecker
    return ag.lp.SphericalSersic(intensity=1.0, effective_radius=2.0, sersic_index=2.0)


def make_lp_1():
    # noinspection PyTypeChecker
    return ag.lp.SphericalSersic(intensity=2.0, effective_radius=2.0, sersic_index=2.0)


def make_mp_0():
    # noinspection PyTypeChecker
    return ag.mp.SphericalIsothermal(einstein_radius=1.0)


def make_mp_1():
    # noinspection PyTypeChecker
    return ag.mp.SphericalIsothermal(einstein_radius=2.0)


def make_lmp_0():
    return ag.lmp.EllipticalSersicRadialGradient()


def make_dmp_0():
    # noinspection PyTypeChecker
    return ag.mp.SphericalNFW(kappa_s=1.0)


def make_dmp_1():
    # noinspection PyTypeChecker
    return ag.mp.SphericalNFW(kappa_s=2.0)


def make_smp_0():
    # noinspection PyTypeChecker
    return ag.lmp.EllipticalSersic(intensity=1.0, mass_to_light_ratio=1.0)


def make_smp_1():
    # noinspection PyTypeChecker
    return ag.lmp.EllipticalSersic(intensity=2.0, mass_to_light_ratio=2.0)


# GALAXY #


def make_gal_x1_lp():
    return ag.Galaxy(redshift=0.5, light_profile_0=make_lp_0())


def make_gal_x2_lp():
    return ag.Galaxy(
        redshift=0.5, light_profile_0=make_lp_0(), light_profile_1=make_lp_1()
    )


def make_gal_x1_mp():
    return ag.Galaxy(redshift=0.5, mass_profile_0=make_mp_0())


def make_gal_x2_mp():
    return ag.Galaxy(
        redshift=0.5, mass_profile_0=make_mp_0(), mass_profile_1=make_mp_1()
    )


def make_gal_x1_lp_x1_mp():
    return ag.Galaxy(
        redshift=0.5, light_profile_0=make_lp_0(), mass_profile_0=make_mp_0()
    )


def make_hyper_galaxy():
    return ag.HyperGalaxy(noise_factor=1.0, noise_power=1.0, contribution_factor=1.0)


# Plane #


def make_plane_7x7():
    return ag.Plane(galaxies=[make_gal_x1_lp_x1_mp()])


def make_plane_x2_galaxy_inversion_7x7():
    source_gal_inversion = ag.Galaxy(
        redshift=1.0,
        pixelization=ag.pix.Rectangular(),
        regularization=ag.reg.Constant(),
    )

    return ag.Plane(galaxies=[make_gal_x1_lp(), source_gal_inversion])


# GALAXY DATA #


def make_gal_data_7x7():
    return ag.GalaxyData(
        image=make_image_7x7(),
        noise_map=make_noise_map_7x7(),
        pixel_scales=make_image_7x7().pixel_scales,
    )


def make_gal_fit_data_7x7_image():
    return ag.MaskedGalaxyDataset(
        galaxy_data=make_gal_data_7x7(), mask=make_sub_mask_7x7(), use_image=True
    )


def make_gal_fit_data_7x7_convergence():
    return ag.MaskedGalaxyDataset(
        galaxy_data=make_gal_data_7x7(), mask=make_sub_mask_7x7(), use_convergence=True
    )


def make_gal_fit_data_7x7_potential():
    return ag.MaskedGalaxyDataset(
        galaxy_data=make_gal_data_7x7(), mask=make_sub_mask_7x7(), use_potential=True
    )


def make_gal_fit_data_7x7_deflections_y():
    return ag.MaskedGalaxyDataset(
        galaxy_data=make_gal_data_7x7(),
        mask=make_sub_mask_7x7(),
        use_deflections_y=True,
    )


def make_gal_fit_data_7x7_deflections_x():
    return ag.MaskedGalaxyDataset(
        galaxy_data=make_gal_data_7x7(),
        mask=make_sub_mask_7x7(),
        use_deflections_x=True,
    )


# GALAXY FIT #


def make_gal_fit_7x7_image():
    return ag.FitGalaxy(
        masked_galaxy_dataset=make_gal_fit_data_7x7_image(),
        model_galaxies=[make_gal_x1_lp()],
    )


def make_gal_fit_7x7_convergence():
    return ag.FitGalaxy(
        masked_galaxy_dataset=make_gal_fit_data_7x7_convergence(),
        model_galaxies=[make_gal_x1_mp()],
    )


def make_gal_fit_7x7_potential():
    return ag.FitGalaxy(
        masked_galaxy_dataset=make_gal_fit_data_7x7_potential(),
        model_galaxies=[make_gal_x1_mp()],
    )


def make_gal_fit_7x7_deflections_y():
    return ag.FitGalaxy(
        masked_galaxy_dataset=make_gal_fit_data_7x7_deflections_y(),
        model_galaxies=[make_gal_x1_mp()],
    )


def make_gal_fit_7x7_deflections_x():
    return ag.FitGalaxy(
        masked_galaxy_dataset=make_gal_fit_data_7x7_deflections_x(),
        model_galaxies=[make_gal_x1_mp()],
    )


# HYPER GALAXIES #


def make_hyper_model_image_7x7():
    return ag.Array.manual_mask(
        np.full(fill_value=5.0, shape=make_mask_7x7().pixels_in_mask),
        mask=make_mask_7x7(),
    )


def make_hyper_galaxy_image_0_7x7():
    return ag.Array.manual_mask(
        np.full(fill_value=2.0, shape=make_mask_7x7().pixels_in_mask),
        mask=make_mask_7x7(),
    )


def make_hyper_galaxy_image_path_dict_7x7():
    hyper_galaxy_image_path_dict = {
        ("g0",): make_hyper_galaxy_image_0_7x7(),
        ("g1",): make_hyper_galaxy_image_1_7x7(),
    }

    return hyper_galaxy_image_path_dict


def make_hyper_galaxy_image_1_7x7():
    return ag.Array.manual_mask(
        np.full(fill_value=3.0, shape=make_mask_7x7().pixels_in_mask),
        mask=make_mask_7x7(),
    )


def make_masked_imaging_fit_7x7():
    return ag.FitImaging(
        masked_imaging=make_masked_imaging_7x7(), plane=make_plane_7x7()
    )


def make_masked_imaging_fit_x2_galaxy_7x7():
    plane = ag.Plane(galaxies=[make_gal_x1_lp(), make_gal_x1_lp()])

    return ag.FitImaging(masked_imaging=make_masked_imaging_7x7(), plane=plane)


def make_masked_imaging_fit_x2_galaxy_inversion_7x7():
    return ag.FitImaging(
        masked_imaging=make_masked_imaging_7x7(),
        plane=make_plane_x2_galaxy_inversion_7x7(),
    )


def make_masked_interferometer_fit_7x7():
    return ag.FitInterferometer(
        masked_interferometer=make_masked_interferometer_7(), plane=make_plane_7x7()
    )


def make_masked_interferometer_fit_x2_galaxy_inversion_7x7():
    return ag.FitInterferometer(
        masked_interferometer=make_masked_interferometer_7(),
        plane=make_plane_x2_galaxy_inversion_7x7(),
    )


def make_samples_with_result():
    galaxies = [
        ag.Galaxy(redshift=0.5, light=ag.lp.EllipticalSersic(intensity=1.0)),
        ag.Galaxy(redshift=1.0, light=ag.lp.EllipticalSersic(intensity=2.0)),
    ]

    plane = ag.Plane(galaxies=galaxies)

    return MockSamples(max_log_likelihood_instance=plane)


def make_phase_data():
    return PhaseDataset(
        settings=ag.SettingsPhaseImaging(), search=MockSearch(name="test_phase")
    )


def make_phase_imaging_7x7():
    return ag.PhaseImaging(search=MockSearch(name="test_phase"))


def make_phase_interferometer_7():
    return ag.PhaseInterferometer(
        search=MockSearch(name="test_phase"), real_space_mask=make_mask_7x7()
    )


def make_include_all():
    return Include(
        origin=True,
        mask=True,
        grid=True,
        border=True,
        positions=True,
        light_profile_centres=True,
        mass_profile_centres=True,
        critical_curves=True,
        caustics=True,
        multiple_images=True,
        inversion_pixelization_grid=True,
        inversion_grid=True,
        inversion_border=True,
        inversion_image_pixelization_grid=True,
        preloaded_critical_curves=ag.GridCoordinates([(1.0, 1.0), (2.0, 2.0)]),
        preload_caustics=ag.GridCoordinates([(1.0, 1.0), (2.0, 2.0)]),
    )



import autofit as af
import autogalaxy as ag
from autogalaxy import exc

import pytest


class TestSetupHyper:
    def test__hyper_searches(self):
        setup = ag.SetupHyper(hyper_galaxies=False)
        assert setup.hyper_galaxies_search == None

        setup = ag.SetupHyper(hyper_galaxies=True)
        assert setup.hyper_galaxies_search.n_live_points == 75
        assert setup.hyper_galaxies_search.evidence_tolerance == pytest.approx(
            0.084, 1.0e-4
        )

        setup = ag.SetupHyper(
            hyper_galaxies=True,
            hyper_galaxies_search=af.DynestyStatic(n_live_points=51),
        )
        assert setup.hyper_galaxies_search.n_live_points == 51
        assert setup.hyper_galaxies_search.evidence_tolerance == pytest.approx(
            0.06, 1.0e-4
        )

        setup = ag.SetupHyper(inversion_search=None)
        assert setup.inversion_search.n_live_points == 50
        assert setup.inversion_search.evidence_tolerance == pytest.approx(0.059, 1.0e-4)

        setup = ag.SetupHyper(inversion_search=af.DynestyStatic(n_live_points=51))
        assert setup.inversion_search.n_live_points == 51

        setup = ag.SetupHyper(hyper_combined_search=af.DynestyStatic(n_live_points=51))
        assert setup.hyper_combined_search.n_live_points == 51

        setup = ag.SetupHyper(hyper_galaxies=True, evidence_tolerance=0.5)
        assert setup.hyper_galaxies_search.evidence_tolerance == pytest.approx(
            0.084, 1.0e-4
        )
        assert setup.inversion_search.evidence_tolerance == 0.5
        assert setup.hyper_combined_search.evidence_tolerance == 0.5

        with pytest.raises(exc.PipelineException):
            ag.SetupHyper(
                hyper_galaxies=True,
                hyper_galaxies_search=af.DynestyStatic(n_live_points=51),
                evidence_tolerance=0.5,
            )

        with pytest.raises(exc.PipelineException):
            ag.SetupHyper(
                inversion_search=af.DynestyStatic(n_live_points=51),
                evidence_tolerance=3.0,
            )

        with pytest.raises(exc.PipelineException):
            ag.SetupHyper(
                hyper_combined_search=af.DynestyStatic(n_live_points=51),
                evidence_tolerance=3.0,
            )

    def test__hyper_galaxies_tag(self):
        setup = ag.SetupHyper(hyper_galaxies=False)
        assert setup.hyper_galaxies_tag == ""

        setup = ag.SetupHyper(hyper_galaxies=True)
        assert setup.hyper_galaxies_tag == "galaxies"

    def test__hyper_image_sky_tag(self):
        setup = ag.SetupHyper(hyper_image_sky=False)
        assert setup.hyper_galaxies_tag == ""

        setup = ag.SetupHyper(hyper_image_sky=True)
        assert setup.hyper_image_sky_tag == "__bg_sky"

    def test__hyper_background_noise_tag(self):
        setup = ag.SetupHyper(hyper_background_noise=False)
        assert setup.hyper_galaxies_tag == ""

        setup = ag.SetupHyper(hyper_background_noise=True)
        assert setup.hyper_background_noise_tag == "__bg_noise"

    def test__tag(self):

        setup = ag.SetupHyper(
            hyper_galaxies=False, hyper_image_sky=False, hyper_background_noise=False
        )

        assert setup.tag == ""

        setup = ag.SetupHyper(
            hyper_galaxies=True, hyper_image_sky=True, hyper_background_noise=True
        )

        assert setup.tag == "hyper[galaxies__bg_sky__bg_noise]"

        setup = ag.SetupHyper(hyper_galaxies=True, hyper_background_noise=True)

        assert setup.tag == "hyper[galaxies__bg_noise]"


class TestSetupLightParametric:
    def test__prior_models_and_tags(self):

        setup = ag.SetupLightParametric(
            bulge_prior_model=af.PriorModel(ag.lp.EllipticalSersic),
            disk_prior_model=af.PriorModel(ag.lp.EllipticalExponential),
        )

        assert setup.bulge_prior_model.cls is ag.lp.EllipticalSersic
        assert setup.disk_prior_model.cls is ag.lp.EllipticalExponential
        assert setup.envelope_prior_model is None

        assert setup.bulge_prior_model_tag == "__bulge_sersic"
        assert setup.disk_prior_model_tag == "__disk_exp"
        assert setup.envelope_prior_model_tag == ""

        setup = ag.SetupLightParametric(
            bulge_prior_model=af.PriorModel(ag.lp.SphericalSersic),
            disk_prior_model=af.PriorModel(ag.lp.SphericalExponential),
            envelope_prior_model=af.PriorModel(ag.lp.EllipticalCoreSersic),
        )

        assert setup.bulge_prior_model_tag == "__bulge_sersic_sph"
        assert setup.disk_prior_model_tag == "__disk_exp_sph"
        assert setup.envelope_prior_model_tag == "__envelope_core_sersic"

    def test__set_bulge_disk_prior_model_assertions(self):

        setup = ag.SetupLightParametric(
            bulge_prior_model=af.PriorModel(ag.lp.EllipticalSersic),
            disk_prior_model=af.PriorModel(ag.lp.EllipticalSersic),
        )

        assert isinstance(
            setup.bulge_prior_model._assertions[0], af.GreaterThanLessThanAssertion
        )

        setup = ag.SetupLightParametric(
            bulge_prior_model=af.PriorModel(ag.mp.EllipticalSersic),
            disk_prior_model=af.PriorModel(ag.mp.EllipticalExponential),
        )

        assert setup.bulge_prior_model._assertions == []

    def test__set_chameleon_prior_model_assertions(self):

        setup = ag.SetupLightParametric(
            bulge_prior_model=af.PriorModel(ag.lp.EllipticalChameleon),
            disk_prior_model=af.PriorModel(ag.lp.SphericalChameleon),
            envelope_prior_model=af.PriorModel(ag.lmp.EllipticalChameleon),
        )

        assert isinstance(
            setup.bulge_prior_model._assertions[0], af.GreaterThanLessThanAssertion
        )
        assert isinstance(
            setup.disk_prior_model._assertions[0], af.GreaterThanLessThanAssertion
        )
        assert isinstance(
            setup.envelope_prior_model._assertions[0], af.GreaterThanLessThanAssertion
        )

        setup = ag.SetupLightParametric(
            bulge_prior_model=af.PriorModel(ag.mp.EllipticalChameleon),
            disk_prior_model=af.PriorModel(ag.mp.SphericalChameleon),
            envelope_prior_model=af.PriorModel(ag.lmp.SphericalChameleon),
        )

        assert isinstance(
            setup.bulge_prior_model._assertions[0], af.GreaterThanLessThanAssertion
        )
        assert isinstance(
            setup.disk_prior_model._assertions[0], af.GreaterThanLessThanAssertion
        )
        assert isinstance(
            setup.envelope_prior_model._assertions[0], af.GreaterThanLessThanAssertion
        )

    def test__input_light_centre__centres_of_prior_models_are_aligned(self):

        setup = ag.SetupLightParametric(
            bulge_prior_model=af.PriorModel(ag.lp.SphericalSersic),
            disk_prior_model=af.PriorModel(ag.lp.SphericalExponential),
            envelope_prior_model=af.PriorModel(ag.lp.SphericalExponential),
            light_centre=(0.0, 0.0),
        )

        assert setup.bulge_prior_model.centre == (0.0, 0.0)
        assert setup.disk_prior_model.centre == (0.0, 0.0)
        assert setup.envelope_prior_model.centre == (0.0, 0.0)

    def test__input_align_centre_and_elliptical_comps__components_are_aligned(self):

        setup = ag.SetupLightParametric(
            bulge_prior_model=af.PriorModel(ag.lp.EllipticalSersic),
            disk_prior_model=af.PriorModel(ag.lp.EllipticalExponential),
            align_bulge_disk_centre=True,
            align_bulge_disk_elliptical_comps=True,
        )

        assert setup.bulge_prior_model.centre == setup.disk_prior_model.centre
        assert (
            setup.bulge_prior_model.elliptical_comps
            == setup.disk_prior_model.elliptical_comps
        )

        setup = ag.SetupLightParametric(
            bulge_prior_model=af.PriorModel(ag.lp.SphericalSersic),
            disk_prior_model=af.PriorModel(ag.lp.SphericalExponential),
            align_bulge_disk_centre=True,
            align_bulge_disk_elliptical_comps=True,
        )

        assert setup.bulge_prior_model.centre == setup.disk_prior_model.centre

    def test__light_centre_tag(self):

        setup = ag.SetupLightParametric(light_centre=None)
        assert setup.light_centre_tag == ""
        setup = ag.SetupLightParametric(light_centre=(2.0, 2.0))
        assert setup.light_centre_tag == "__centre_(2.00,2.00)"
        setup = ag.SetupLightParametric(light_centre=(3.0, 4.0))
        assert setup.light_centre_tag == "__centre_(3.00,4.00)"
        setup = ag.SetupLightParametric(light_centre=(3.027, 4.033))
        assert setup.light_centre_tag == "__centre_(3.03,4.03)"

    def test__align_bulge_disk_tags(self):
        light = ag.SetupLightParametric(align_bulge_disk_centre=False)
        assert light.align_bulge_disk_centre_tag == ""
        light = ag.SetupLightParametric(align_bulge_disk_centre=True)
        assert light.align_bulge_disk_centre_tag == "_centre"

        light = ag.SetupLightParametric(align_bulge_disk_elliptical_comps=False)
        assert light.align_bulge_disk_elliptical_comps_tag == ""
        light = ag.SetupLightParametric(align_bulge_disk_elliptical_comps=True)
        assert light.align_bulge_disk_elliptical_comps_tag == "_ell"

    def test__tag(self):

        setup = ag.SetupLightParametric(light_centre=None)
        assert (
            setup.tag
            == "light[parametric__bulge_sersic__disk_exp__align_bulge_disk_centre]"
        )

        setup = ag.SetupLightParametric(
            light_centre=(3.027, 4.033), align_bulge_disk_centre=False
        )
        assert (
            setup.tag == "light[parametric__bulge_sersic__disk_exp__centre_(3.03,4.03)]"
        )

        light = ag.SetupLightParametric(
            align_bulge_disk_centre=True, align_bulge_disk_elliptical_comps=True
        )

        assert (
            light.tag
            == "light[parametric__bulge_sersic__disk_exp__align_bulge_disk_centre_ell]"
        )

        light = ag.SetupLightParametric(
            bulge_prior_model=af.PriorModel(ag.lp.SphericalSersic),
            disk_prior_model=af.PriorModel(ag.lp.SphericalExponential),
            envelope_prior_model=af.PriorModel(ag.lp.EllipticalCoreSersic),
            align_bulge_disk_centre=True,
            align_bulge_disk_elliptical_comps=True,
        )

        assert (
            light.tag
            == "light[parametric__bulge_sersic_sph__disk_exp_sph__envelope_core_sersic__align_bulge_disk_centre_ell]"
        )


class TestSetupLightInversion:
    def test__pixelization_prior_model__model_depends_on_inversion_pixels_fixed(self):

        setup = ag.SetupLightInversion(
            pixelization_prior_model=af.PriorModel(ag.pix.Rectangular),
            regularization_prior_model=af.PriorModel(ag.reg.Regularization),
        )

        assert setup.pixelization_prior_model.cls is ag.pix.Rectangular

        setup = ag.SetupLightInversion(
            pixelization_prior_model=af.PriorModel(ag.pix.VoronoiBrightnessImage),
            regularization_prior_model=af.PriorModel(ag.reg.Regularization),
        )

        assert setup.pixelization_prior_model.cls is ag.pix.VoronoiBrightnessImage

        setup = ag.SetupLightInversion(
            pixelization_prior_model=af.PriorModel(ag.pix.VoronoiBrightnessImage),
            inversion_pixels_fixed=100,
            regularization_prior_model=af.PriorModel(ag.reg.Regularization),
        )

        assert isinstance(setup.pixelization_prior_model, af.PriorModel)
        assert setup.pixelization_prior_model.pixels == 100

    def test__pixelization_tag(self):

        setup = ag.SetupLightInversion(
            pixelization_prior_model=af.PriorModel(ag.pix.Rectangular),
            regularization_prior_model=af.PriorModel(ag.reg.Regularization),
        )
        assert setup.pixelization_tag == "__pix_rect"

        setup = ag.SetupLightInversion(
            pixelization_prior_model=af.PriorModel(ag.pix.VoronoiBrightnessImage),
            regularization_prior_model=af.PriorModel(ag.reg.Regularization),
        )
        assert setup.pixelization_tag == "__pix_voro_image"

    def test__regularization_tag(self):

        setup = ag.SetupLightInversion(
            pixelization_prior_model=af.PriorModel(ag.pix.Pixelization),
            regularization_prior_model=af.PriorModel(ag.reg.Constant),
        )
        assert setup.regularization_tag == "__reg_const"

        setup = ag.SetupLightInversion(
            pixelization_prior_model=af.PriorModel(ag.pix.Pixelization),
            regularization_prior_model=af.PriorModel(ag.reg.AdaptiveBrightness),
        )
        assert setup.regularization_tag == "__reg_adapt_bright"

    def test__inversion_pixels_fixed_tag(self):

        setup = ag.SetupLightInversion(
            inversion_pixels_fixed=None,
            pixelization_prior_model=af.PriorModel(ag.pix.Pixelization),
            regularization_prior_model=af.PriorModel(ag.reg.AdaptiveBrightness),
        )
        assert setup.inversion_pixels_fixed_tag == ""

        setup = ag.SetupLightInversion(
            inversion_pixels_fixed=100,
            pixelization_prior_model=af.PriorModel(ag.pix.Pixelization),
            regularization_prior_model=af.PriorModel(ag.reg.AdaptiveBrightness),
        )
        assert setup.inversion_pixels_fixed_tag == ""

        setup = ag.SetupLightInversion(
            inversion_pixels_fixed=100,
            pixelization_prior_model=af.PriorModel(ag.pix.VoronoiBrightnessImage),
            regularization_prior_model=af.PriorModel(ag.reg.AdaptiveBrightness),
        )
        assert setup.inversion_pixels_fixed_tag == "_100"

    def test__tag(self):

        setup = ag.SetupLightInversion(
            pixelization_prior_model=af.PriorModel(ag.pix.Rectangular),
            regularization_prior_model=af.PriorModel(ag.reg.Constant),
            inversion_pixels_fixed=100,
        )

        assert setup.tag == "light[inversion__pix_rect__reg_const]"

        setup = ag.SetupLightInversion(
            pixelization_prior_model=af.PriorModel(ag.pix.VoronoiBrightnessImage),
            regularization_prior_model=af.PriorModel(ag.reg.AdaptiveBrightness),
            inversion_pixels_fixed=None,
        )

        assert setup.tag == "light[inversion__pix_voro_image__reg_adapt_bright]"

        setup = ag.SetupLightInversion(
            pixelization_prior_model=af.PriorModel(ag.pix.VoronoiBrightnessImage),
            regularization_prior_model=af.PriorModel(ag.reg.AdaptiveBrightness),
            inversion_pixels_fixed=100,
        )

        assert setup.tag == "light[inversion__pix_voro_image_100__reg_adapt_bright]"


class TestAbstractSetupMass:
    def test__align_centre_to_mass_centre(self):

        mass = af.PriorModel(ag.mp.SphericalIsothermal)

        source = ag.SetupMassTotal(mass_centre=(1.0, 2.0))

        mass = source.align_centre_to_mass_centre(mass_prior_model=mass)

        assert mass.centre == (1.0, 2.0)

    def test__unfix_mass_centre(self):

        mass = af.PriorModel(ag.mp.SphericalIsothermal)
        mass.centre = (1.0, 2.0)

        setup_mass = ag.SetupMassTotal()

        mass = af.PriorModel(ag.mp.SphericalIsothermal)

        mass = setup_mass.unfix_mass_centre(mass_prior_model=mass)

        # assert mass.centre.centre_0.mean == 5.0
        # assert mass.centre.centre_0.sigma == 0.05
        # assert mass.centre.centre_1.mean == 6.0
        # assert mass.centre.centre_1.sigma == 0.05


class TestSetupMassTotal:
    def test__mass_prior_model_and_tags(self):

        setup = ag.SetupMassTotal(
            mass_prior_model=af.PriorModel(ag.mp.EllipticalPowerLaw)
        )

        assert setup.mass_prior_model.cls is ag.mp.EllipticalPowerLaw
        assert setup.mass_prior_model_tag == "__power_law"

        setup = ag.SetupMassTotal(
            mass_prior_model=af.PriorModel(ag.mp.EllipticalIsothermal)
        )

        assert setup.mass_prior_model.cls is ag.mp.EllipticalIsothermal
        assert setup.mass_prior_model_tag == "__sie"

    def test__mass_centre_updates_mass_prior_model(self):

        setup = ag.SetupMassTotal(
            mass_prior_model=af.PriorModel(ag.mp.EllipticalIsothermal),
            mass_centre=(0.0, 1.0),
        )

        assert setup.mass_prior_model.centre == (0.0, 1.0)

    def test__mass_centre_tag(self):
        setup = ag.SetupMassTotal(mass_centre=None)
        assert setup.mass_centre_tag == ""
        setup = ag.SetupMassTotal(mass_centre=(2.0, 2.0))
        assert setup.mass_centre_tag == "__centre_(2.00,2.00)"
        setup = ag.SetupMassTotal(mass_centre=(3.0, 4.0))
        assert setup.mass_centre_tag == "__centre_(3.00,4.00)"
        setup = ag.SetupMassTotal(mass_centre=(3.027, 4.033))
        assert setup.mass_centre_tag == "__centre_(3.03,4.03)"

    def test__align_centre_of_mass_to_light(self):

        mass = af.PriorModel(ag.mp.SphericalIsothermal)

        source = ag.SetupMassTotal(align_bulge_mass_centre=False)

        mass = source.align_centre_of_mass_to_light(
            mass_prior_model=mass, light_centre=(1.0, 2.0)
        )

        assert mass.centre.centre_0.mean == 1.0
        assert mass.centre.centre_0.sigma == 0.1
        assert mass.centre.centre_0.mean == 1.0
        assert mass.centre.centre_0.sigma == 0.1

        source = ag.SetupMassTotal(align_bulge_mass_centre=True)

        mass = source.align_centre_of_mass_to_light(
            mass_prior_model=mass, light_centre=(1.0, 2.0)
        )

        assert mass.centre == (1.0, 2.0)

    def test__tag(self):

        setup = ag.SetupMassTotal(mass_centre=None)
        assert setup.tag == "mass[total__power_law]"

        setup = ag.SetupMassTotal(mass_centre=(3.027, 4.033))
        assert setup.tag == "mass[total__power_law__centre_(3.03,4.03)]"

        setup = ag.SetupMassTotal(
            mass_centre=(3.027, 4.033),
            mass_prior_model=af.PriorModel(ag.mp.EllipticalIsothermal),
        )
        assert setup.tag == "mass[total__sie__centre_(3.03,4.03)]"


class TestSetupMassLightDark:
    def test__prior_models_and_tags(self):

        setup = ag.SetupMassLightDark(
            bulge_prior_model=af.PriorModel(ag.lmp.EllipticalSersic),
            disk_prior_model=af.PriorModel(ag.lmp.EllipticalExponential),
        )

        assert setup.bulge_prior_model.cls is ag.lmp.EllipticalSersic
        assert setup.disk_prior_model.cls is ag.lmp.EllipticalExponential
        assert setup.envelope_prior_model is None
        assert setup.dark_prior_model.cls is ag.mp.SphericalNFWMCRLudlow

        assert setup.bulge_prior_model_tag == "__bulge_sersic"
        assert setup.disk_prior_model_tag == "__disk_exp"
        assert setup.envelope_prior_model_tag == ""
        assert setup.dark_prior_model_tag == "__dark_nfw_sph_ludlow"

        setup = ag.SetupMassLightDark(
            bulge_prior_model=af.PriorModel(ag.lmp.EllipticalDevVaucouleurs),
            disk_prior_model=af.PriorModel(ag.lmp.SphericalDevVaucouleurs),
            envelope_prior_model=af.PriorModel(ag.lmp.SphericalExponential),
            dark_prior_model=af.PriorModel(ag.mp.EllipticalNFW),
        )

        assert setup.bulge_prior_model.cls is ag.lmp.EllipticalDevVaucouleurs
        assert setup.disk_prior_model.cls is ag.lmp.SphericalDevVaucouleurs
        assert setup.envelope_prior_model.cls is ag.lmp.SphericalExponential
        assert setup.dark_prior_model.cls is ag.mp.EllipticalNFW

        assert setup.bulge_prior_model_tag == "__bulge_dev"
        assert setup.disk_prior_model_tag == "__disk_dev_sph"
        assert setup.envelope_prior_model_tag == "__envelope_exp_sph"
        assert setup.dark_prior_model_tag == "__dark_nfw"

    def test__set_chameleon_prior_model_assertions(self):

        setup = ag.SetupMassLightDark(
            bulge_prior_model=af.PriorModel(ag.lp.EllipticalChameleon),
            disk_prior_model=af.PriorModel(ag.lp.SphericalChameleon),
            envelope_prior_model=af.PriorModel(ag.lmp.EllipticalChameleon),
        )

        assert isinstance(
            setup.bulge_prior_model._assertions[0], af.GreaterThanLessThanAssertion
        )
        assert isinstance(
            setup.disk_prior_model._assertions[0], af.GreaterThanLessThanAssertion
        )
        assert isinstance(
            setup.envelope_prior_model._assertions[0], af.GreaterThanLessThanAssertion
        )

        setup = ag.SetupMassLightDark(
            bulge_prior_model=af.PriorModel(ag.mp.EllipticalChameleon),
            disk_prior_model=af.PriorModel(ag.mp.SphericalChameleon),
            envelope_prior_model=af.PriorModel(ag.lmp.SphericalChameleon),
        )

        assert isinstance(
            setup.bulge_prior_model._assertions[0], af.GreaterThanLessThanAssertion
        )
        assert isinstance(
            setup.disk_prior_model._assertions[0], af.GreaterThanLessThanAssertion
        )
        assert isinstance(
            setup.envelope_prior_model._assertions[0], af.GreaterThanLessThanAssertion
        )

    def test__consstant_mass_to_light_ratio__sets_mass_to_light_ratios_of_light_and_mass_profiles(
        self
    ):

        setup = ag.SetupMassLightDark(
            bulge_prior_model=ag.lmp.EllipticalSersic,
            disk_prior_model=ag.lmp.EllipticalSersic,
            envelope_prior_model=ag.lmp.EllipticalSersic,
            constant_mass_to_light_ratio=False,
        )

        assert (
            setup.bulge_prior_model.mass_to_light_ratio
            != setup.disk_prior_model.mass_to_light_ratio
        )
        assert (
            setup.bulge_prior_model.mass_to_light_ratio
            != setup.envelope_prior_model.mass_to_light_ratio
        )
        assert (
            setup.disk_prior_model.mass_to_light_ratio
            != setup.envelope_prior_model.mass_to_light_ratio
        )

        setup = ag.SetupMassLightDark(
            bulge_prior_model=ag.lmp.EllipticalSersic,
            disk_prior_model=ag.lmp.EllipticalSersic,
            envelope_prior_model=ag.lmp.EllipticalSersic,
            constant_mass_to_light_ratio=True,
        )

        assert (
            setup.bulge_prior_model.mass_to_light_ratio
            == setup.disk_prior_model.mass_to_light_ratio
        )
        assert (
            setup.bulge_prior_model.mass_to_light_ratio
            == setup.envelope_prior_model.mass_to_light_ratio
        )
        assert (
            setup.disk_prior_model.mass_to_light_ratio
            == setup.envelope_prior_model.mass_to_light_ratio
        )

        setup = ag.SetupMassLightDark(
            bulge_prior_model=ag.lmp.EllipticalSersic,
            disk_prior_model=ag.lmp.EllipticalSersic,
            envelope_prior_model=None,
            constant_mass_to_light_ratio=True,
        )

        assert (
            setup.bulge_prior_model.mass_to_light_ratio
            == setup.disk_prior_model.mass_to_light_ratio
        )

        setup = ag.SetupMassLightDark(
            bulge_prior_model=ag.lmp.EllipticalSersic,
            disk_prior_model=None,
            envelope_prior_model=None,
            constant_mass_to_light_ratio=True,
        )

        assert (
            setup.bulge_prior_model.mass_to_light_ratio
            == setup.bulge_prior_model.mass_to_light_ratio
        )

    def test__constant_mass_to_light_ratio_tag(self):

        setup = ag.SetupMassLightDark(constant_mass_to_light_ratio=True)
        assert setup.constant_mass_to_light_ratio_tag == "__mlr_const"

        setup = ag.SetupMassLightDark(constant_mass_to_light_ratio=False)
        assert setup.constant_mass_to_light_ratio_tag == "__mlr_free"

    def test__align_bulge_dark_tag(self):

        setup = ag.SetupMassLightDark(align_bulge_dark_centre=False)
        assert setup.align_bulge_dark_centre_tag == ""

        setup = ag.SetupMassLightDark(align_bulge_dark_centre=True)
        assert setup.align_bulge_dark_centre_tag == "__align_bulge_dark_centre"

    def test__tag(self):

        setup = ag.SetupMassLightDark(
            bulge_prior_model=af.PriorModel(ag.lmp.EllipticalSersic),
            disk_prior_model=af.PriorModel(ag.lmp.EllipticalExponential),
            constant_mass_to_light_ratio=True,
            align_bulge_dark_centre=True,
        )
        assert (
            setup.tag
            == "mass[light_dark__bulge_sersic__disk_exp__mlr_const__dark_nfw_sph_ludlow__align_bulge_dark_centre]"
        )


class TestSMBH:
    def test__tag(self):

        setup = ag.SetupSMBH(smbh_centre_fixed=True)

        assert setup.tag == "smbh[point_mass__centre_fixed]"

        setup = ag.SetupSMBH(smbh_centre_fixed=False)

        assert setup.tag == "smbh[point_mass__centre_free]"

    def test__smbh_from_centre(self):

        setup = ag.SetupSMBH(smbh_centre_fixed=True)
        smbh = setup.smbh_from_centre(centre=(0.0, 0.0))

        assert isinstance(smbh, af.PriorModel)
        assert smbh.centre == (0.0, 0.0)

        setup = ag.SetupSMBH(smbh_centre_fixed=False)
        smbh = setup.smbh_from_centre(centre=(0.1, 0.2), centre_sigma=0.2)

        assert isinstance(smbh, af.PriorModel)

        assert isinstance(smbh.centre[0], af.GaussianPrior)
        assert smbh.centre[0].mean == 0.1
        assert smbh.centre[0].sigma == 0.2

        assert isinstance(smbh.centre[1], af.GaussianPrior)
        assert smbh.centre[1].mean == 0.2
        assert smbh.centre[1].sigma == 0.2


class TestSetupPipeline:
    def test__setup__passes_light_setup_to_mass_light_setup(self):

        setup_light = ag.SetupLightParametric(
            bulge_prior_model=af.PriorModel(ag.lp.EllipticalSersic),
            disk_prior_model=af.PriorModel(ag.lp.EllipticalExponential),
        )

        setup_mass = ag.SetupMassLightDark()

        setup_pipeline = ag.SetupPipeline(
            setup_light=setup_light, setup_mass=setup_mass
        )

        assert (
            setup_pipeline.setup_light.bulge_prior_model.cls is ag.lp.EllipticalSersic
        )
        assert (
            setup_pipeline.setup_light.disk_prior_model.cls
            is ag.lp.EllipticalExponential
        )
        assert setup_pipeline.setup_light.envelope_prior_model is None

        assert setup_pipeline.setup_mass.bulge_prior_model.cls is ag.lp.EllipticalSersic
        assert (
            setup_pipeline.setup_mass.disk_prior_model.cls
            is ag.lp.EllipticalExponential
        )
        assert setup_pipeline.setup_mass.envelope_prior_model is None

    def test__tag(self):

        setup_light = ag.SetupLightParametric(
            light_centre=(1.0, 2.0), align_bulge_disk_centre=False
        )

        setup = ag.SetupPipeline(setup_light=setup_light)

        assert (
            setup.tag == "setup__"
            "light[parametric__bulge_sersic__disk_exp__centre_(1.00,2.00)]"
        )

        setup_hyper = ag.SetupHyper(hyper_galaxies=True, hyper_background_noise=True)

        setup_light = ag.SetupLightParametric(light_centre=(1.0, 2.0))

        setup = ag.SetupPipeline(setup_hyper=setup_hyper, setup_light=setup_light)

        assert (
            setup.tag == "setup__hyper[galaxies__bg_noise]__"
            "light[parametric__bulge_sersic__disk_exp__align_bulge_disk_centre__centre_(1.00,2.00)]"
        )

        smbh = ag.SetupSMBH(smbh_centre_fixed=True)

        setup = ag.SetupPipeline(setup_smbh=smbh)

        assert setup.tag == "setup__smbh[point_mass__centre_fixed]"
